unit Rectangles;
{ (c) 2006, 2008, Kees Hemerik, Tom Verhoeff (TU/e) }
{ ADTs for axis-parallel grid rectangles and list thereof }

interface

uses
  Contnrs;

type
  // An object of class TRectangle represents a rectangle. It has:
  // - primitive queries XL, YL, XH, YH for inspecting the coordinates
  // - derived queries for inspecting width and height
  // - some additional functions for checking well-formedness and for checking
  //   whether the rectangle contains a given point.
  // - some commands for changing one or all coordinates
  // - An invariant WellFormed

  { TRectangle }

  TRectangle =
  class(TObject)
  private
    // fields
    FXL: Integer;
    FYL: Integer;
    FXH: Integer;
    FYH: Integer;
  public
    // construction ------------------------------------------------------------
    constructor Create(AXL, AYL, AXH, AYH: Integer);
    // pre: ValidCoordinates(AXL, AYL, AXH, AYH)
    // post: XL = AXL, YL = AYL, XH = AXH, YH = AYH

    // primitive queries -------------------------------------------------------
    function XL: Integer;
    function YL: Integer;
    function XH: Integer;
    function YH: Integer;

    // derived queries ---------------------------------------------------------
    function Width: Integer;
    // pre: true
    // ret: XH - XL + 1
    function Height: Integer;
    // pre: true
    // ret: YH - YL + 1

    function Contains(AX, AY: Integer): Boolean;
    // pre: true
    // ret: (XL <= AX) and (AX <= XH) and (YL <= AY) and (AY <= YH)

    function ValidCoordinates(AXL, AYL, AXH, AYH: Integer): Boolean;
    // pre: true
    // ret: (AXL <= AXH) and (AYL <= AYH)

    function WellFormed: Boolean;
    // pre: true
    // ret: ValidCoordinates(XL, YL, XH, YH)

    // commands -----------------------------------------------------
    procedure SetXL(AXL: Integer);
    // pre: AXL <= XH
    // effect: XL := AXL

    procedure SetYL(AYL: Integer);
    // pre: AYL <= YH
    // effect: YL := AYL

    procedure SetXH(AXH: Integer);
    // pre: XL <= AXH
    // effect: XH := AXH

    procedure SetYH(AYH: Integer);
    // pre: YL <= AYH
    // effect: YH := AYH

    procedure Shift(AX, AY: Integer);
    // pre: true
    // effect: XL, YL, XH, YH := XL + AX, YL + AY, XH + AX, YH + AY

    procedure Hull(const ARectangle: TRectangle);
    // pre: true
    // post: Self = smallest rectangle enclosing Self and ARectangle

    // invariants ---------------------------------------------------
    // WF: WellFormed
  end;


  // TRectangleList is a class for representing a list of rectangles
  // Its methods have the same functionality as the methods with the same name
  // in the VCL class TObjectList (see Delphi Help), but they operate on a
  // TRectangle rather than a TObject.
  TRectangleList =
  class(TObject)
  protected
    FList: TObjectList;
  public
    // construction and destruction
    constructor Create;
    destructor Destroy; override;

    // queries
    function Count: Integer;
    function GetItem(I: Integer): TRectangle;
    function IndexOf(ARectangle: TRectangle): Integer;
    function First: TRectangle;
    function Last: TRectangle;

    // commands
    procedure Clear;
    procedure Add(ARectangle: TRectangle);
    procedure Insert(I: Integer; ARectangle: TRectangle);
    procedure Delete(I: Integer);
  end;



implementation //===============================================================

uses
  SysUtils, Math;

{ TRectangle }

function TRectangle.Contains(AX, AY: Integer): Boolean;
begin
  Result := (XL <= AX) and (AX <= XH) and (YL <= AY) and (AY <= YH);
end;

constructor TRectangle.Create(AXL, AYL, AXH, AYH: Integer);
begin
  Assert(
    ValidCoordinates(AXL, AYL, AXH, AYH),
    Format('TRectangle.Create.pre failed. Parameter values: %d, %d, %d, %d.', [AXL, AYL, AXH, AYH]));
  inherited Create;
  FXL := AXL;
  FYL := AYL;
  FXH := AXH;
  FYH := AYH;
end;

function TRectangle.Height: Integer;
begin
  Result := YH - YL + 1;
end;

procedure TRectangle.SetXH(AXH: Integer);
begin
  Assert(
    XL <= AXH,
    Format('TRectangle.SetXH.pre failed. XL = %d; AXH = %d.', [XL, AXH]));
  FXH := AXH;
end;

procedure TRectangle.SetXL(AXL: Integer);
begin
  Assert(
    AXL <= XH,
    Format('TRectangle.SetXL.pre failed. AXL = %d; XH = %d.', [AXL, XH]));
  FXL := AXL;
end;

procedure TRectangle.SetYH(AYH: Integer);
begin
  Assert(
    YL <= AYH,
    Format('TRectangle.SetYH.pre failed. YL = %d; AYH = %d.', [YL, AYH]));
  FYH := AYH;
end;

procedure TRectangle.SetYL(AYL: Integer);
begin
  Assert(
    AYL <= YH,
    Format('TRectangle.SetYL.pre failed. AYL = %d; YH = %d.', [AYL, YH]));
  FYL := AYL;
end;

procedure TRectangle.Shift(AX, AY: Integer);
begin
  FXL := FXL + AX;
  FYL := FYL + AY;
  FXH := FXH + AX;
  FYH := FYH + AY;
end;

procedure TRectangle.Hull(const ARectangle: TRectangle);
begin
  FXL := Min ( FXL, ARectangle.XL );
  FYL := Min ( FYL, ARectangle.YL );
  FXH := Max ( FXH, ARectangle.XH );
  FYH := Max ( FYH, ARectangle.YH );
end;

function TRectangle.ValidCoordinates(AXL, AYL, AXH, AYH: Integer): Boolean;
begin
  Result := (AXL <= AXH) and (AYL <= AYH);
end;

function TRectangle.WellFormed: Boolean;
begin
  Result := Validcoordinates(XL, YL, XH, YH);
end;

function TRectangle.Width: Integer;
begin
  Result := XH - XL + 1;
end;

function TRectangle.XH: Integer;
begin
  Result := FXH;
end;

function TRectangle.XL: Integer;
begin
  Result := FXL;
end;

function TRectangle.YH: Integer;
begin
  Result := FYH;
end;

function TRectangle.YL: Integer;
begin
  Result := FYL;
end;

{ TRectangleList }

procedure TRectangleList.Add(ARectangle: TRectangle);
begin
  FList.Add(ARectangle);
end;

procedure TRectangleList.Clear;
begin
  FList.Clear;
end;

function TRectangleList.Count: Integer;
begin
  Result := FList.Count;
end;

constructor TRectangleList.Create;
begin
  inherited Create;
  FList := TObjectList.Create;
end;

destructor TRectangleList.Destroy;
begin
  FList.Free;
  inherited;
end;

function TRectangleList.First: TRectangle;
begin
  Result := FList.First as TRectangle;
end;

function TRectangleList.GetItem(I: Integer): TRectangle;
begin
  Result := FList.Items[I] as TRectangle;
end;

function TRectangleList.IndexOf(ARectangle: TRectangle): Integer;
begin
  Result := FList.IndexOf(ARectangle);
end;

procedure TRectangleList.Insert(I: Integer; ARectangle: TRectangle);
begin
  FList.Insert(I, ARectangle);
end;

function TRectangleList.Last: TRectangle;
begin
  Result := FList.Last as Trectangle;
end;

procedure TRectangleList.Delete(I: Integer);
begin
  FList.Delete(I);
end;

end.
